#include "fit.h"
#include <math.h>
#include <stdio.h>
#include <vector>
#include <algorithm>
#include <assert.h>

//#define DEBUG_OUTPUT

// Helper class for fitting polynomials to points.
class polynomial_point_data
{
public:
    polynomial_point_data(const double *xvalues,
                          const double *yvalues,
                          size_t nvalues)
        : m_xvalues(xvalues)
        , m_yvalues(yvalues)
        , m_nvalues(nvalues)
    {
    }

    const double *m_xvalues, *m_yvalues;
    size_t m_nvalues;
};

// Calculate the sn um squared error and of a polynomial against a 
// set of points. 
double
polynomial_point_error(const void *p,
                       const double *parameters,
                       size_t nparameters)
{
    double error(0.0);

    if(p)
    {
        const polynomial_point_data *data = 
            static_cast<const polynomial_point_data *>(p);

        for(size_t i=0;i<data->m_nvalues;i++)
        {
            double tmp;
            tmp = evaluate_polynomial(data->m_xvalues[i],
                                      parameters,
                                      nparameters);
            tmp-= data->m_yvalues[i];
            error+= (tmp * tmp);
        }
        error/= (double) data->m_nvalues;
    }

    return error;
}

// Iteration functions.
double newton(double dedp, double d2edp2)
{
    return -(dedp/(d2edp2));
}

double lm(double dedp, double d2edp2, double mu)
{
    return -(dedp/(d2edp2 + mu));
}

////////////////////////////////////////////////////////////////////////////////
// Interface

// Write a polynomial in a form where it is easy to just copy into
// a c program or gnuplot.
void 
write_polynomial(const double *parameters, size_t nparameters)
{

    printf("0.0");
    if(parameters && nparameters)
    {
        for(size_t i=0;i<nparameters;i++)
        {
            printf(" + %g", parameters[i]);
            for(size_t j=0;j<i;j++)
            {
                printf("*x");
            }
        }
        putchar('\n');
    }
    else
    {
        fprintf(stderr, "invalid parameters to write_polynomial\n");
    }
}

// Evaluates a polynomial y = a + bx + cx^2 + dx^3...
// with the parameters a,b,c,d given in the parameter vector.
double
evaluate_polynomial(double x, const double *parameters, size_t nparameters)
{
    double result(0.0);

    if(parameters && nparameters)
    {
        result = parameters[0];
    
        double power(x);

        for(size_t i=1;i<nparameters;i++)
        {
            double coefficient;
            coefficient = parameters[i];
    
            result+= coefficient * power;

            power*= x;
        }
    }
    else
    {
        fprintf(stderr, "invalid parameters to evaluate_polynomial\n");
    }
    return result;
}

// Differentiate a polynomial y = a + bx + cx^2 + dx^3...
// with the parameters a,b,c,d given in the parameter vector.
double
evaluate_differential(double x, const double *parameters, size_t nparameters)
{
    double result(0.0);

    if(parameters && nparameters)
    {
        double power(1.0);

        for(size_t i=1;i<nparameters;i++)
        {
            double coefficient, multiplier;
            coefficient = parameters[i];
            multiplier = i;
    
            result+= coefficient * power * multiplier;

            power*= x;
        }
    }
    else
    {
        fprintf(stderr, "invalid parameters to evaluate_polynomial\n");
    }
    return result;
}


// RANSAC helpers.

class ransac_data
{
public:
    ransac_data(
            double (*function)(const void *, const size_t *, size_t, const double *, size_t),
            const void *data,
            const size_t *items,
            const size_t nitems)
        : m_function(function)
        , m_data(data)
        , m_items(items)
        , m_nitems(nitems)
    {
    }

    double (*m_function)(const void *, const size_t *, size_t, const double *, size_t);
    const void *m_data;
    const size_t *m_items;
    const size_t m_nitems;
};

static double ransac_function(const void *p,
                              const double *parameters,
                              size_t nparameters)
{
    const ransac_data *data((const ransac_data *)p);

    if(data && data->m_function && data->m_items && data->m_nitems)
    {
        return data->m_function(data->m_data, data->m_items, data->m_nitems,
                                parameters, nparameters);
    }
    else
    {
        assert(0);
        return 0;
    }
}

// Generic fit function using RANSAC.
// Arguments:
// "function" - an error function, which should return rms error of some
//            subset of the data.
//            parameters are as follows:
//            "data" - void pointer to whatever it needs.
//            "items" - indices of items included in this optimisation pass
//            "nitems" - the number of items in "items"
//            "parameters" - parameters vector to optimise.
//            "nparameters" - size of parameter vector to optimise.
// "parameters"  - parameters vector to optimise.
// "nparameters" - size of parameter vector to optimise.
// "min_error"   - terminate least square step when squared error drops below this
// "convergence" - terminate when change in squared error between steps drops
//                 below this
// "max_steps"   - termination threshold for optimiser
// "max_iterations" - number of ransac iterations.
// Return Value: squared error, or -1 if arguments are invalid.
size_t fit_ransac(
    double (*function)(const void *, const size_t *, size_t, const double *, size_t),
    const void *data,
    size_t nitems,
    double *parameters,
    size_t nparameters,
    double min_error,
    double convergence,
    double max_steps,
    size_t nransac_items,
    double consensus_error,
    size_t iterations)
{
    size_t result(0);

    if(data && nitems && parameters && nparameters &&(nransac_items < nitems))
    {
        size_t *all_items;
        double *best_model, *cur_model;

        all_items = new size_t[nitems];
        best_model = new double[nparameters];
        cur_model = new double[nparameters];

        if(all_items && best_model && cur_model)
        {
            std::vector<size_t> best_consensus;

            for(size_t i=0;i<nitems;i++)
            {
                all_items[i] = i;
            }

            for(size_t i=0;i<iterations;i++)
            {
                std::vector<size_t> cur_consensus;

                // Initialise parameters to seed.
                for(size_t j=0;j<nparameters;j++)
                {
                    cur_model[j] = parameters[j];
                }

                // Pick a set to optimise.
                std::random_shuffle(all_items, all_items + nitems);

                {
                    std::vector<size_t> consensus;

                    ransac_data fit_data(function,
                                         data,
                                         all_items,
                                         nransac_items);

                    fit_parameters(ransac_function,
                                   (const void *)&fit_data,
                                   cur_model,
                                   nparameters,
                                   min_error,
                                   convergence,
                                   max_steps);
                }

                // Check the model
                cur_consensus.clear();
                for(size_t j=0;j<nitems;j++)
                {
                    if(function(data, &j, 1, cur_model, nparameters) <
                       consensus_error)
                    {
                        cur_consensus.push_back(j);
                    }
                }

                if(cur_consensus.size() > best_consensus.size())
                {
#ifdef DEBUG_OUTPUT
                    printf("New consensus.  Size = %li\n", cur_consensus.size());
#endif
                    for(size_t j=0;j<nparameters;j++)
                    {
                        best_model[j] = cur_model[j];
                    }
                    cur_consensus.swap(best_consensus);
                }
            }
            // Override the ransac ids with the best consensus ids, and 
            // optimise against consensus.  Ugly, but saves an alloc.
            for(size_t i=0;i<best_consensus.size();i++)
            {
                all_items[i] = best_consensus[i];
            }

            if(best_consensus.size())
            {
                ransac_data fit_data(function,
                                     data,
                                     all_items,
                                     best_consensus.size());

                fit_parameters(ransac_function,
                               (const void *)&fit_data,
                               best_model,
                               nparameters,
                               min_error,
                               convergence,
                               max_steps);

                // Finally, output data
                for(size_t i=0;i<nparameters;i++)
                {
                    parameters[i] = best_model[i];
                }
                result = best_consensus.size();
            }
        }
        if(all_items)
        {
            delete all_items;
        }
        if(cur_model)
        {
            delete cur_model;
        }
        if(best_model)
        {
            delete best_model;
        }
    }

    return result;
}


double
fit_parameters(double (*function)(const void *, const double *, size_t), 
               const void *data,
               double *parameters,
               size_t nparameters,
               double min_error,
               double convergence,
               size_t max_steps)
{
    double result;
    size_t *p = new size_t[nparameters];
    for(size_t i=0;i<nparameters;i++)
    {
        p[i] = i;
    }
    result = fit_subset(function, data, parameters, nparameters, p, 
                        nparameters, min_error, convergence, max_steps);
    delete[] p;

    return result;
}

double
fit_subset(double (*function)(const void *, const double *, size_t),
           const void *data,
           double *parameters,
           size_t nparameters,
           size_t *used,
           size_t nused,
           double min_error,
           double convergence,
           size_t max_steps)
{
    if(function && data && parameters && nparameters)
    {
        double dp(1.0e-5);
        double dp2(dp*dp);
        double step_size(1.0/sqrt(nparameters));
        double mu(0.3);
        size_t step(0);
    
        bool converged(false);
        double error, prev_error;
        double *dedp,           // 1st derivative of error vs parameters
               *d2edp2,         // 2nd derivative of error vs parameters
               *tmp_parameters, // temporaty parameter vector
               *update;         // vector of updates for this step
    
        tmp_parameters        = new double[nparameters];
        dedp                  = new double[nparameters];
        d2edp2                = new double[nparameters];
        update                = new double[nparameters];
    
        error = function(data, parameters, nparameters);
    
        prev_error = error;
    
        do
        {
            for(size_t i=0;i<nparameters;i++)
            {
                tmp_parameters[i] = parameters[i];
            }
    
            // Calculate derivatives & second derivatives
            for(size_t i=0;i<nused;i++)
            {
                double ep,    // error at p
                       eppdp, // error at p+dp
                       epmdp; // error at p-dp
    
                ep = error;
    
                // Value if we add step to parameter.
                tmp_parameters[used[i]] = parameters[used[i]] + dp;

                eppdp = function(data, tmp_parameters, nparameters);
    
                // Value if we subtract step from parameter.
                tmp_parameters[used[i]] = parameters[used[i]] - dp;

                epmdp = function(data, tmp_parameters, nparameters);
    
                // Restore parameter value.
                tmp_parameters[used[i]] = parameters[used[i]];
            
                // Get derivatives.
                dedp[used[i]] = (eppdp - epmdp) / (dp * 2);
                d2edp2[used[i]] = ((eppdp + epmdp) - (2.0 * ep)) / dp2;
            }
    
            // update
            for(size_t i=0;i<nused;i++)
            {
                //update[i] = newton(dedp[i], d2edp2[i]);
                update[used[i]] = lm(dedp[used[i]], d2edp2[used[i]], mu);
    
                //printf("i = %i, newton = %g, lm = %g\n",
                //       i, 
                //       update[i],
                //       lm(dedp[i], d2edp2[i], mu));
    
    
                update[used[i]]*= step_size;
            }
    
            for(size_t i=0;i<nused;i++)
            {
                parameters[used[i]]+= update[used[i]]; 
            }

    
            prev_error = error;

            error = function(data, parameters, nparameters);
    
            if(error > prev_error)
            {
                // error increasing.  Damp.
                step_size*= 0.75;
#ifdef DEBUG_OUTPUT
                fprintf(stderr, "updated step_size to %g\n", step_size);
#endif
            }
            else
            {
                if((fabs((prev_error - error)) < convergence) || 
                   (step_size < convergence) ||
                   (error < min_error))
                {
                    converged = true;
                } 
            }
    
#ifdef DEBUG_OUTPUT
            if((step % 500) == 0)
            {
                fprintf(stderr, "Error at step %ld = %g\n", step, error);
            }
#endif

            step++;
        } while(!converged && (step < max_steps));
    
    
        delete[] tmp_parameters;
        delete[] dedp;
        delete[] d2edp2;
        delete[] update;
#ifdef DEBUG_OUTPUT 
        fprintf(stderr, "terminated after %ld steps\n", step);
#endif
        return error;
    }
    else
    {
#ifdef DEBUG_OUTPUT
        fprintf(stderr, "Invalid arguments\n");
#endif
        return -1;
    }
}

// Wrap up the fit_parameters function for simple polynomial point fitting.
// 
double
fit_polynomial_to_points(const double *xvalues,
                         const double *yvalues,
                         size_t nvalues,
                         double *parameters,
                         size_t nparameters,
                         double min_error,
                         double convergence,
                         size_t max_steps)
{
    polynomial_point_data data(xvalues, yvalues, nvalues);

    return fit_parameters(polynomial_point_error, 
                          (const void *)&data, 
                          parameters,
                          nparameters,
                          min_error,
                          convergence,
                          max_steps);
}
