#include <math.h>
#include <string.h>
#include "image_types.h"
#include "integral_histogram.h"


namespace s3d_merge
{

////////////////////////////////////////////////////////////////////////////////

IntegralHistogram::IntegralHistogram(uint32_t bins, const Image8 &img,
                                     uint32_t(*pixel_to_bin)(uint32_t,
                                                             uint32_t,
                                                             const Image8 *,
                                                             void *),
                                     void *data)

    : image<uint32_t>(img.get_width(), img.get_height(), bins)
{
    BuildHistogram(bins, img, 0, 0, img.get_width(), img.get_height(),
                   pixel_to_bin, data);
}

IntegralHistogram::IntegralHistogram(uint32_t bins, const Image8 &img,
                                     uint32_t x0, uint32_t y0,
                                     uint32_t width, uint32_t height,
                                     uint32_t(*pixel_to_bin)(uint32_t,
                                                             uint32_t,
                                                             const Image8 *,
                                                             void *),
                                     void *data)
    : image<uint32_t>(width, height, bins)
{
    BuildHistogram(bins, img, x0, y0, width, height, pixel_to_bin, data);
}

IntegralHistogram::~IntegralHistogram()
{
}

void
IntegralHistogram::BuildHistogram(uint32_t bins, const Image8 &img,
                                  uint32_t x0, uint32_t y0,
                                  uint32_t width, uint32_t height,
                                  uint32_t(*pixel_to_bin)(uint32_t,
                                                          uint32_t,
                                                          const Image8 *,
                                                          void *),
                                  void *data)
{
    uint32_t x,y;
    size_t index(0);

    memset(get_buffer(), 0, 
           get_width() * get_height() * bins * sizeof(uint32_t));

    for(y=0;y<get_height();y++)
    {
        for(x=0;x<get_width();x++)
        {
            uint32_t bin;

            // Add 1 to this bin if we are in range.
            if((x < width) && (y < height))
            {
                bin = pixel_to_bin(x + x0, y + y0, &img, data);

                if(bin < bins)
                {
                    get_buffer()[index + size_t(bin)] = 1;
                }
            }

            if(x)
            {
                // Add previous pixel in x
                const uint32_t *pixel;
                uint32_t i;
                pixel = get_pixel(x-1, y);
                for(i=0;i<bins;i++)
                {
                    get_buffer()[index + i]+= pixel[i];
                }
            }
            if(y)
            {
                // Add previous pixel in y
                const uint32_t *pixel;
                uint32_t i;

                pixel = get_pixel(x, y-1);
                for(i=0;i<bins;i++)
                {
                    get_buffer()[index + i]+= pixel[i];
                }
            }
            if(x && y)
            {
                // subtract previous pixel in x,y
                const uint32_t *pixel;
                uint32_t i;

                pixel = get_pixel(x-1, y-1);
                for(i=0;i<bins;i++)
                {
                    get_buffer()[index + i]-= pixel[i];
                }
            }

            index+= bins;
        }
    }
}

void 
IntegralHistogram::GetHistogram(uint32_t x1, uint32_t y1,
                                uint32_t x2, uint32_t y2,
                                uint32_t *result) const
{
    uint32_t i;

    if(x1 > x2)
    {
        uint32_t tmp = x1;
        x1 = x2;
        x2 = tmp;
    }
    if(y1 > y2)
    {
        uint32_t tmp = y1;
        y1 = y2;
        y2 = tmp;
    }

    for(i=0;i<get_channels();i++)
    {
        result[i] = 0;
    }

    if((x1 < get_width()) && (y1 < get_height()))
    {
        const uint32_t *a, *b, *c, *d;
        //
        //  a-----b
        //  |     |
        //  d-----c
        //
        // We want to get inclusive points, so we move the top left up, 
        // potentially making a, b, c invalid in the process.

        a = b = c = d = NULL;

        if(x2 >= get_width())
        {
            x2 = get_width() - 1;
        }
        if(y2 >= get_height())
        {
            y2 = get_height() - 1;
        }

        c = get_pixel(x2, y2);

        if(x1 && y1)
        {
            a = get_pixel(x1 - 1, y1 - 1);
        }
        if(x1)
        {
            d = get_pixel(x1 - 1, y2);
        }
        if(y1)
        {
            b = get_pixel(x2, y1 - 1);
        }

        // calculate histogram.
        // TODO: optimise
        for(i=0;i<get_channels();i++)
        {
            if(a) result[i]+=a[i];
            if(c) result[i]+=c[i];
            if(b) result[i]-=b[i];
            if(d) result[i]-=d[i];
        }
    }
}

}
