#ifndef __KD_TREE_H__
#define __KD_TREE_H__

#include <stdint.h>
#include <cstdlib>
#include <cmath>
#include <vector>
#include <algorithm>
#include <functional>
#include <limits>
#include <utility>
#include <cassert>

#ifdef KD_TREE_DEBUG
#include <iostream>
#endif

namespace spectral {

template <typename ScalarType = float,
          size_t Dimension = 3>
class Point
{
public:
    typedef ScalarType        scalar_type;

    // Construction
    Point()
    {
        for(size_t i=0;i<Dimension;i++)
        {
            m_point[i] = 0;
        }
    }

    Point(const Point &other)
    {
        *this = other;
    }

    // Operators
    const Point &operator=(const Point<ScalarType,Dimension> &other)
    {
        for(size_t i=0;i<Dimension;i++)
        {
            m_point[i] = other[i];
        }
        return *this;
    }

    scalar_type &operator[](size_t idx)
    {
#ifndef _NDEBUG
        assert(idx < Dimension);
#endif
        return m_point[idx];
    }

    const scalar_type &operator[](size_t idx) const
    {
#ifndef _NDEBUG
        assert(idx < Dimension);
#endif
        return m_point[idx];
    }

    // Arithmetic operators
    Point operator+(const Point &other) const
    {
        Point result;
        for(size_t idx=0;idx<Dimension;idx++)
        {
            result[idx] = m_point[idx]+other[idx];
        }
        return result;
    }

    const Point &operator+=(const Point &other)
    {
        for(size_t idx=0;idx<Dimension;idx++)
        {
            m_point[idx]+= other[idx];
        }
        return *this;
    }

    Point operator-(const Point &other) const
    {
        Point result;
        for(size_t idx=0;idx<Dimension;idx++)
        {
            result[idx] = m_point[idx]-other[idx];
        }
        return result;
    }

    const Point &operator-=(const Point &other)
    {
        for(size_t idx=0;idx<Dimension;idx++)
        {
            m_point[idx]-= other[idx];
        }
        return *this;
    }

    Point operator*(scalar_type scale) const
    {
        Point result;
        for(size_t idx=0;idx<Dimension;idx++)
        {
            result[idx] = m_point[idx]*scale;
        }
        return result;
    }

    const Point &operator*=(scalar_type scale)
    {
        for(size_t idx=0;idx<Dimension;idx++)
        {
            m_point[idx]*= scale;
        }
        return *this;
    }

    Point operator/(scalar_type scale) const
    {
        Point result;
        for(size_t idx=0;idx<Dimension;idx++)
        {
            result[idx] = m_point[idx]/scale;
        }
        return result;
    }

    const Point &operator/=(scalar_type scale)
    {
        for(size_t idx=0;idx<Dimension;idx++)
        {
            m_point[idx]/= scale;
        }
        return *this;
    }

    void Normalise(void)
    {
        scalar_type len(0);

        for(size_t idx=0;idx<Dimension;idx++)
        {
            len+=(m_point[idx]* m_point[idx]);
        }

        len = sqrt(len);

        for(size_t idx=0;idx<Dimension;idx++)
        {
            m_point[idx]/= len;
        }
    }

    // statics

    static scalar_type DotProduct(Point p1, Point p2)
    {
        scalar_type result(0);
        for(size_t idx=0;idx<Dimension;idx++)
        {
            result+= (p1[idx] * p2[idx]);
        }
        return result;
    }

    static scalar_type SqDistance(Point p1, Point p2)
    {
        scalar_type result(0);
        for(size_t idx=0;idx<Dimension;idx++)
        {
            scalar_type tmp;
            tmp = (p1[idx] - p2[idx]);
            tmp*= tmp;
            result+= tmp;
        }
        return result;
    }

    static scalar_type Distance(Point p1, Point p2)
    {
        return sqrt(SqDistance(p1, p2));
    }
private:
    scalar_type m_point[Dimension];
};

template<typename ScalarType = float,
         size_t Dimension = 3,
         typename IndexType = size_t>
class KDTree
{
public:
    typedef ScalarType                          scalar_type;
    typedef IndexType                           index_type;
    typedef Point<ScalarType, Dimension>        point_type;

private:
    // Tree node representation.
    class Node
    {
    public:
       Node(point_type position,
            index_type index)
           : m_position(position)
           , m_index(index)
           , m_child_count(0)
           , m_plane(0)
       {
       }

       Node(point_type position,
            index_type index,
            index_type child_count,
            size_t plane)
           : m_position(position)
           , m_index(index)
           , m_child_count(child_count)
           , m_plane(plane)
       {
       }

       point_type          m_position;
       index_type          m_index;
       index_type          m_child_count;
       size_t              m_plane;
    };

    std::vector<Node> m_tree;

public:
    // Constructors.
    KDTree()
    {
    }

    template <class PointIterator>
    KDTree(PointIterator begin, 
           PointIterator end)
    {
    
        if(begin != end)
        {
            typename std::vector<std::pair<point_type, index_type> > tmp_points;
            PointIterator it(begin);
            index_type idx(0);

            m_tree.reserve(end - begin);
            tmp_points.reserve(m_tree.size());

            while(it != end)
            {
                tmp_points.push_back(std::make_pair(*it, idx));
                idx++;
                it++;
            }
            BuildTree(tmp_points.begin(), tmp_points.end());
        }
    }

    // Queries

    // Brute force reference implementation of FindItemsInRange.  TODO: Remove later
    void FindItemsInRangeBF(point_type min, point_type max, typename std::vector<index_type> &out) const
    {
        for(size_t i=0;i<m_tree.size();i++)
        {
            bool found;
            found = true;
   
            for(size_t j=0;j<Dimension;j++)
            {
                if((m_tree[i].m_position[j] < min[j]) ||
                   (m_tree[i].m_position[j] > max[j]))
                {
                    found = false;
                    break;
                }
            }
            if(found)
            {
                out.push_back(m_tree[i].m_index);
            }
        }
    }

    // Find all items in the axis aligned hypercube bounded by min,max.
    void FindItemsInRange(point_type min, point_type max, typename std::vector<index_type> &out) const
    {
        if(m_tree.size())
        {
            std::vector<size_t> stack;
            for(size_t i=0;i<Dimension;i++)
            {
                if(max[i] < min[i])
                {
                    std::swap(max[i], min[i]);
                }
            }
    
            stack.push_back(0); // root
    
            while(stack.size())
            {
                size_t next_idx;
    
                // Process whatever is at the top of the stack.
                next_idx = stack.back();
                stack.pop_back();            
    
                {
                    const Node &node(m_tree[next_idx]);
                    bool traverse_left(true), traverse_right(true);
    
                    if(node.m_position[node.m_plane] < min[node.m_plane])
                    {
                        traverse_left = false;
                    }
    
                    if(node.m_position[node.m_plane] > max[node.m_plane])
                    {
                        traverse_right = false;
                    }
    
                    if(traverse_left && HasLeftChild(next_idx))
                    {
                        stack.push_back(GetLeftChild(next_idx));
                    }
    
                    if(traverse_right && HasRightChild(next_idx))
                    {
                        stack.push_back(GetRightChild(next_idx));
                    }
    
                    if(traverse_left && traverse_right)
                    {
                        bool found(true);
   
                        for(size_t i=0;i<Dimension;i++)
                        {
                            if((node.m_position[i] < min[i]) ||
                               (node.m_position[i] > max[i]))
                            {
                                found = false;
                                break;
                            }
                        }
                        if(found)
                        {
                            out.push_back(node.m_index);
                        }
                    }
                }
            }
        }
    }

    // Brute force reference implementation of FindClosestItem
    bool FindClosestItemBF(point_type p, index_type &out, scalar_type &distance) const
    {
        bool found(false);

        if(m_tree.size())
        {
            scalar_type candidate(0);
            scalar_type best_distance = std::numeric_limits<scalar_type>::max();

            for(size_t i=0;i<m_tree.size();i++)
            {
                scalar_type cur_distance;

                cur_distance = point_type::Distance(p, m_tree[i].m_position);

                if(cur_distance < best_distance)
                {
                    candidate = m_tree[i].m_index;
                    best_distance = cur_distance;
                    found = true;
                }
            }

            if(found)
            {
                out = candidate;
                distance = best_distance;
            }
        }
        return found;
    }

    // Find the closest item to a point.
    bool FindClosestItem(point_type p, index_type &out, scalar_type &distance) const
    {
        bool found(false);

        if(m_tree.size())
        {
            std::vector<size_t> stack;
            scalar_type candidate(0);
            scalar_type best_distance = std::numeric_limits<scalar_type>::max();
            stack.push_back(0);
          
            while(stack.size())
            {
                size_t next_idx;
    
                // Process whatever is at the top of the stack.
                next_idx = stack.back();
                stack.pop_back();            

                {
                    const Node &node(m_tree[next_idx]);
                    bool traverse_left(true), traverse_right(true);
                    scalar_type cur_distance;

                    cur_distance = point_type::Distance(node.m_position, p);

                    if(cur_distance < best_distance)
                    {
                        best_distance = cur_distance;
                        candidate = node.m_index;
                        found = true;
                    }

                    if(node.m_position[node.m_plane] < (p[node.m_plane] - best_distance))
                    {
                        traverse_left = false;
                    }
                    if(node.m_position[node.m_plane] > (p[node.m_plane] + best_distance))
                    {
                        traverse_right = false;
                    }

                    if(traverse_left && HasLeftChild(next_idx))
                    {
                        stack.push_back(GetLeftChild(next_idx));
                    }
    
                    if(traverse_right && HasRightChild(next_idx))
                    {
                        stack.push_back(GetRightChild(next_idx));
                    }
                }
            }
            if(found)
            {
                out = candidate;
                distance = best_distance;
            }
        }

        return found;
    }

    // Find the closest item to a point.
    void FindClosestItems(point_type p, size_t num, std::vector<index_type> &out) const
    {
        scalar_type cull_distance = std::numeric_limits<scalar_type>::max();
        out.clear();

        if(m_tree.size() && num)
        {
            std::vector<size_t> stack;
            std::vector<std::pair<scalar_type, index_type> > candidates;
            stack.push_back(0);
          
            while(stack.size())
            {
                size_t next_idx;
    
                // Process whatever is at the top of the stack.
                next_idx = stack.back();
                stack.pop_back();            

                {
                    const Node &node(m_tree[next_idx]);
                    bool traverse_left(true), traverse_right(true);
                    scalar_type cur_distance;

                    cur_distance = point_type::Distance(node.m_position, p);

                    if(cur_distance <= cull_distance)
                    {
                        candidates.push_back(std::make_pair(cur_distance, node.m_index));

                        if(candidates.size() > num)
                        {
                            SearchComparator cmp;

                            std::sort(candidates.begin(), candidates.end(), cmp);
                            candidates.pop_back();

                            cull_distance = candidates.back().first;
                        }
                    }

                    if(node.m_position[node.m_plane] < (p[node.m_plane] - cull_distance))
                    {
                        traverse_left = false;
                    }
                    if(node.m_position[node.m_plane] > (p[node.m_plane] + cull_distance))
                    {
                        traverse_right = false;
                    }

                    if(traverse_left && HasLeftChild(next_idx))
                    {
                        stack.push_back(GetLeftChild(next_idx));
                    }
    
                    if(traverse_right && HasRightChild(next_idx))
                    {
                        stack.push_back(GetRightChild(next_idx));
                    }
                }
            }
            if(candidates.size())
            {
                SearchComparator cmp;
                std::sort(candidates.begin(), candidates.end(), cmp);

                out.reserve(candidates.size());
                for(size_t i=0;i<candidates.size();i++)
                {
                    out.push_back(candidates[i].second);
                }
            }
        }
    }

private:
    bool HasLeftChild(size_t pos) const
    {
        return(m_tree[pos].m_child_count != 0);
    }
  
    size_t GetLeftChild(size_t pos) const
    {
        if(m_tree[pos].m_child_count)
        {
            return pos + 1;
        }
        else
        {
            assert(0);
            return pos;
        }
    }

    bool HasRightChild(size_t pos) const
    {
        if(HasLeftChild(pos))
        {
            return (m_tree[pos].m_child_count > 
                    (m_tree[pos+1].m_child_count + 1));
        }
        return false;
    }

    size_t GetRightChild(size_t pos) const
    {
        if(HasRightChild(pos))
        {
            // step over the entire left subtree.
            return pos + (m_tree[pos+1].m_child_count) + 2;
        }
        else
        {
            assert(0);
            return pos;
        }
    }

    // Helper classes

    // Build a kd tree from a vector of points and indices.  We do this by
    //
    void BuildTree(
        typename std::vector<std::pair<point_type, index_type> >::iterator begin,
        typename std::vector<std::pair<point_type, index_type> >::iterator end)
    {
        index_type size(end - begin);

        if(size)
        {
            if(size == 1)
            {
                // Leaf node.
#ifdef KD_TREE_DEBUG
                std::cout << "Pushed leaf node " << begin->second << std::endl;
#endif
                m_tree.push_back(Node(begin->first, begin->second));
            }
            else
            {
                // Decide which direction to split on.
                scalar_type mean[Dimension];
                scalar_type sum_variance[Dimension];

                for(size_t i=0;i<Dimension;i++)
                {
                    mean[i] = 0;
                    sum_variance[i] = 0;
                }

                // First get mean in each dimension.
                for(typename std::vector<std::pair<point_type, index_type> >::iterator it(begin);
                    it != end; 
                    it++)
                {
                    for(size_t i=0;i<Dimension;i++)
                    {
                        mean[i]+= (it->first)[i];
                    }
                }

                for(size_t i=0;i<Dimension;i++)
                {
                    mean[i]/= size;
                }

                // Now calculate the sum of variances.
                for(typename std::vector<std::pair<point_type, index_type> >::iterator it(begin);
                    it != end; 
                    it++)
                {
                    for(size_t i=0;i<Dimension;i++)
                    {
                        scalar_type val;
                        val = (it->first)[i] - mean[i];
                        val*= val;
                        sum_variance[i]+= val;
                    }
                }

                // pick the axis of largest variance.
                {
                    size_t split_axis(0);

                    scalar_type max(0);

                    for(size_t i=0;i<Dimension;i++)
                    {
                        if(sum_variance[i] > max)
                        {
                            max = sum_variance[i];
                            split_axis = i;
                        }
                    }

                    // Sort along split axis.
                    SplitComparator cmp(split_axis);

                    typename std::vector<std::pair<point_type, index_type> >::iterator 
                        pivot, pp1;

                    // True is below.
                    std::sort(begin, end, cmp);

                    {
                        size_t distance = (end - begin);
                        distance/= 2;
                        if(!distance) distance = 1;

                        pivot = begin + distance;
                    }
                    pp1 = pivot;
                    pp1++;

                    // Push this.
#ifdef KD_TREE_DEBUG
                    std::cout << "Pushed internal node " << pivot->second << " with " << (size - 1) << " children" << std::endl;
#endif
                    m_tree.push_back(Node(pivot->first,
                                          pivot->second,
                                          size-1,
                                          split_axis));


//#ifndef _NDEBUG
                    // If we have a right subtree we must have a left one.
                    if((begin == pivot) && (pp1 != end))
                    {
                        assert(0);
                    }
//#endif

                    // Build left sub tree.
                    // All items in left subtree are < this,
#ifdef KD_TREE_DEBUG
                    std::cout << "left" << std::endl;
#endif
                    BuildTree(begin, pivot);

                    // Build right sub tree.
                    // All items in right subtree are >= this,
#ifdef KD_TREE_DEBUG
                    std::cout << "right" << std::endl;
#endif
                    BuildTree(pp1, end);

#ifdef KD_TREE_DEBUG
                    std::cout << "up" << std::endl;
#endif
                }
            }
        }
        else
        {
            // Nothing to do.
        }
    }
    
    class SplitComparator : public std::binary_function<std::pair<point_type, index_type>, std::pair<point_type, index_type>, bool>
    {
    public:
        SplitComparator(size_t axis)
            : m_axis(axis)
        {
        }

        bool operator()(std::pair<point_type, index_type> p1,
                        std::pair<point_type, index_type> p2) const
        {
            // We want to bias toward the left, as the structure we are using
            // relies on the existence of left subtrees always.
            return p1.first[m_axis] < p2.first[m_axis];
        }

        size_t m_axis;
    };

    class SearchComparator : public std::binary_function<std::pair<scalar_type, index_type>, std::pair<scalar_type, index_type>, bool>
    {
    public:

        bool operator()(std::pair<scalar_type, index_type> i1,
                        std::pair<scalar_type, index_type> i2) const
        {
            return i1.first < i2.first;
        }
    };
};

}

#endif
