#include "merge.h"

#include <string>
#include <vector>
#include <cstdlib>
#include <cstdio>

#include "image_io.h"
#include "image_types.h"
#include "image_data.h"
#include "merge_utils.h"

#include "align.h"

s3d_merge::FloatImage *
s3d_merge::MergeNoAlign(const std::vector<std::string> &image_filenames,
                        const FloatImage *dark_frame,
                        s3d_merge::MergeType type)
{
    FloatImage *result(NULL);

    if(image_filenames.size())
    {
        s3d_merge::Merge *accumulated_output;

        if(type == MERGE_AVERAGE)
        {
            accumulated_output = new s3d_merge::Average(dark_frame);
        }
        else
        {
            accumulated_output = new s3d_merge::Lighten(dark_frame);
        }

        if(image_filenames.size() && accumulated_output)
        {
            for(size_t i=0;i<image_filenames.size();i++)
            {
                s3d_merge::ImageData *id;

                id = s3d_merge::LoadImageData(image_filenames[i].c_str());

                if(id)
                {
                    if(id->Image())
                    {
                        if(accumulated_output->AddFrame(id->Image()))
                        {
                            printf("Frame \"%s\" added.\n", image_filenames[i].c_str());
                        }
                        else
                        {
                            printf("Frame \"%s\" rejected.\n", image_filenames[i].c_str());
                        }
                    }
                    delete id;
                }
                else
                {
                    printf("Unable to load Frame \"%s\".\n", image_filenames[i].c_str());
                }
            }

            printf("Post processing\n");

            if(accumulated_output)
            {
                accumulated_output->PostProcess();
            }
        }
        // Write output.
        if(accumulated_output->GetImage())
        {
            result = new FloatImage(*(accumulated_output->GetImage()));
        }

        // Clean up
        if(accumulated_output)
        {
            delete accumulated_output;
        }
    }

    return result;
}

s3d_merge::FloatImage *
s3d_merge::MergeAlignStars(const std::vector<std::string> &image_filenames,
                           const FloatImage *dark_frame,
                           s3d_merge::MergeType /*type*/)
{
    FloatImage *result(NULL);

    if(image_filenames.size())
    {
        s3d_merge::AverageAlign* accumulated_output(NULL);

        if(image_filenames.size())
        {
            KDTree2D *ref_tree(NULL);
            std::vector<Point2D> ref_points;
            PatternTree *ref_pattern_tree(NULL);
            std::vector<PatternNode> ref_patterns;

            double alignment[ALIGN_NUM_PARAMETERS];
            for(size_t i=0;i<size_t(ALIGN_NUM_PARAMETERS);i++)
            {
                alignment[i] = 0.0;
            }

            for(size_t i=0;i<image_filenames.size();i++)
            {
                s3d_merge::ImageData *id;

                id = s3d_merge::LoadImageData(image_filenames[i].c_str());

                if(id)
                {
                    if(id->Image())
                    {
                        std::vector<StarRef> stars;

                        if(!accumulated_output)
                        {
                            accumulated_output = new s3d_merge::AverageAlign(dark_frame, id->Image());
                        }

                        printf("Image %li - \"%s\"\n", i, image_filenames[i].c_str());
                        FindStars(id->Image(), stars);

                        if(stars.size() > 4)
                        {

                            if(ref_tree)
                            {
                                std::vector<Point2D> cur_points;
                                for(size_t j=0;j<stars.size();j++)
                                {
                                    cur_points.push_back(stars[j].GetPoint());
                                }

                                // Align image.
                                printf("Fitting %li stars to reference image\n", stars.size());

                                if(FitStars(ref_pattern_tree,
                                            ref_patterns,
                                            ref_tree,
                                            ref_points,
                                            cur_points,
                                            alignment))
                                {
                                    printf("Rotate %lf degrees around (0,0), Translate (%lf, %lf)\n",
                                           alignment[0], alignment[1], alignment[2]);

                                    
                                    if(accumulated_output->AddFrame(id->Image(), 
                                                                    alignment[ALIGN_ANGLE],
                                                                    alignment[ALIGN_TRANSLATE_X],
                                                                    alignment[ALIGN_TRANSLATE_Y]))
                                    {
                                    }
                                }
                            }
                            else
                            {
                                printf("Found %li stars in reference image\n", stars.size());
    
                                for(size_t j=0;j<stars.size();j++)
                                {
                                    ref_points.push_back(stars[j].GetPoint());
                                }
                                ref_tree = new KDTree2D(ref_points.begin(), ref_points.end());
                                GeneratePatternNodes(ref_points, ref_tree, ref_patterns);
                                ref_pattern_tree = new PatternTree(ref_patterns.begin(), ref_patterns.end());
                            }
                        }
/*
                        if(accumulated_output->AddFrame(id->Image()))
                        {
                            printf("Frame \"%s\" added.\n", image_filenames[i].c_str());
                        }
                        else
                        {
                            printf("Frame \"%s\" rejected.\n", image_filenames[i].c_str());
                        }
*/
                    }
                    delete id;
                }
                else
                {
                    printf("Unable to load Frame \"%s\".\n", image_filenames[i].c_str());
                }
            }

            printf("Post processing\n");

            if(accumulated_output)
            {
                accumulated_output->PostProcess();
            }
            if(ref_tree)
            {
                delete ref_tree;
            }
            if(ref_pattern_tree)
            {
                delete ref_pattern_tree;
            }

        }
        // Write output.
        if(accumulated_output->GetImage())
        {
            result = new FloatImage(*(accumulated_output->GetImage()));
        }

        // Clean up
        if(accumulated_output)
        {
            delete accumulated_output;
        }
    }

    return result;
}

