#include "merge_utils.h"
#include "image_types.h"
#include "align.h"
#include "sampler.h"
static const double one_eighty_over_pi(57.2957795131); // degrees to radians


// a = _c(_a, _b)
#define CRUNCH(_a, _b, _c)                                        \
{                                                                 \
    float *afp = _a->get_buffer();                                \
    uint16_t *bip = _b->get_buffer();                             \
    size_t items = _b->get_width() * _b->get_height() * 3;        \
                                                                  \
                                                                  \
    for(size_t i=0;i<items;i++)                                   \
    {                                                             \
        float val;                                                \
        val = (float(*bip) / 65536.0);                            \
        *afp = _c(*afp, val);                                     \
        afp++;                                                    \
        bip++;                                                    \
    }                                                             \
}

// a = _d(_a, _b - c)
#define CRUNCHD(_a, _b, _c, _d)                                   \
{                                                                 \
    float *afp = _a->get_buffer();                                \
    uint16_t *bip = _b->get_buffer();                             \
    const float *cfp = _c->get_buffer();                          \
    size_t items = _b->get_width() * _b->get_height() * 3;        \
                                                                  \
                                                                  \
    for(size_t i=0;i<items;i++)                                   \
    {                                                             \
        float val;                                                \
        val = (float(*bip) / 65536.0);                            \
        val-= *cfp;                                               \
        *afp = _d(*afp, val);                                     \
        afp++;                                                    \
        bip++;                                                    \
        cfp++;                                                    \
    }                                                             \
}

inline float px_set_value(float, float in)
{
    return in;
}

inline float px_add_value(float cur, float in)
{
    return in + cur;
}

inline float px_max_value(float cur, float in)
{
    return (in > cur)?in:cur;
}

s3d_merge::Merge::Merge()
    : m_image(NULL)
    , m_dark(0)
{
}

s3d_merge::Merge::Merge(const FloatImage *dark)
    : m_image(NULL)
    , m_dark(dark)
{
}

s3d_merge::Merge::~Merge()
{
    if(m_image)
    {
        delete m_image;
    }
}


bool
s3d_merge::Merge::SetImage(const Image16 *img)
{
    bool result(false);

    if(m_image)
    {
        delete m_image;
        m_image = NULL;
    }

    if(img)
    {
        m_image = new s3d_merge::FloatImage(
            img->get_width(),
            img->get_height(),
            3);

        if((m_dark) &&
           (m_dark->get_width() == m_image->get_width()) &&
           (m_dark->get_height() == m_image->get_height()))
        {
            CRUNCHD(m_image, img, m_dark, px_set_value);
            result = true;
        }
        else
        {
            CRUNCH(m_image, img, px_set_value);
            result = true;
        }
    }

    return result;
}


////////////////////////////////////////////////////////////////////////////////

s3d_merge::Average::Average()
    : Merge()
    , m_count(0)
{
}

s3d_merge::Average::Average(const FloatImage *dark)
    : Merge(dark)
    , m_count(0)
{
}

s3d_merge::Average::~Average()
{
}

bool
s3d_merge::Average::AddFrame(const Image16 *img)
{
    bool added(false);

    if(img)
    {
        if(GetImage())
        {
            // Check image.
            if((img->get_width() == GetImage()->get_width()) &&
               (img->get_height() == GetImage()->get_height()))
            {
                // Bonza.
                if((GetDark()) &&
                   (GetDark()->get_width() == GetImage()->get_width()) &&
                   (GetDark()->get_height() == GetImage()->get_height()))
                {
                    CRUNCHD(GetImage(), img, GetDark(), px_add_value);
                    added = true;
                }
                else
                {
                    CRUNCH((GetImage()), img, px_add_value);
                    added = true;
                }
                m_count++;
            }
        }
        else
        {
            // Add image
            m_count = 1;
            return SetImage(img);
        }
    }
    return added;
}

bool 
s3d_merge::Average::PostProcess(void)
{
    if(GetImage() && (m_count != 0.0))
    {
        float *p = GetImage()->get_buffer();
        size_t size = GetImage()->get_width() * GetImage()->get_height() * 3;
        for(size_t i=0;i<size;i++)
        {
            *p/= m_count;
            p++;
        }
    }
    return true;
}

////////////////////////////////////////////////////////////////////////////////

s3d_merge::Lighten::Lighten()
    : Merge()
{
}

s3d_merge::Lighten::Lighten(const FloatImage *dark)
    : Merge(dark)
{
}

s3d_merge::Lighten::~Lighten()
{
}

bool
s3d_merge::Lighten::AddFrame(const Image16 *img)
{
    bool added(false);

    if(img)
    {
        if(GetImage())
        {
            // Check image.
            if((img->get_width() == GetImage()->get_width()) &&
               (img->get_height() == GetImage()->get_height()))
            {
                // Bonza.
                if((GetDark()) &&
                   (GetDark()->get_width() == GetImage()->get_width()) &&
                   (GetDark()->get_height() == GetImage()->get_height()))
                {
                    CRUNCHD(GetImage(), img, GetDark(), px_max_value);
                    added = true;
                }
                else
                {
                    CRUNCH((GetImage()), img, px_max_value);
                    added = true;
                }
            }
        }
        else
        {
            // Add image
            return SetImage(img);
        }
    }
    return added;
}

bool 
s3d_merge::Lighten::PostProcess(void)
{
    return true;
}
////////////////////////////////////////////////////////////////////////////////

// Inverse of transform in align.  TODO: Generic transform class!
static s3d_merge::Point2D InvTransformPoint(const s3d_merge::Point2D &point, double angle, double x, double y)
{
    s3d_merge::Point2D tmp_point, use_point, i, j, t;

    double cos_a(cos(-angle / one_eighty_over_pi));
    double sin_a(sin(-angle / one_eighty_over_pi));

    t[0] = x;
    t[1] = y;

    use_point = point - t;

    i[0] =  cos_a;
    i[1] =  sin_a;

    j[0] = -sin_a;
    j[1] =  cos_a;

    tmp_point = ((i * use_point[0]) + (j * use_point[1]));

    return tmp_point;
}


bool 
s3d_merge::AverageAlign::AddFrame(const s3d_merge::Image16 *img, double angle, double tx, double ty)
{
    bool added(false);

    if(img && m_image)
    {
        FloatImage *tmp_img = MakeFloatImage(img);

        Point2D i, j;
        {
            double cos_a(cos(-angle / one_eighty_over_pi));
            double sin_a(sin(-angle / one_eighty_over_pi));

            i[0] =  cos_a;
            i[1] =  sin_a;

            j[0] = -sin_a;
            j[1] =  cos_a;
        }
  
        if(tmp_img)
        {
            float *p = m_image->get_buffer();
            float *wp = m_weights->get_buffer();

            LanczosSampler sampler(ImageSampler::CLAMP, ImageSampler::CLAMP, 2);

            Point2D point, tpoint;
            float bg[3], col[3];
            bg[0] = bg[1] = bg[2] = 0.0;

            for(uint32_t y=0;y<m_image->get_height();y++)
            {
                point[1] = y - ty;
                if(y % 200 == 0)
                {
                    printf("Remapping image %f percent complete\n", float(y*100) / float(m_image->get_height()));
                }
                for(uint32_t x=0;x<m_image->get_width();x++)
                {
                    point[0] = x - tx;
                   
                    tpoint = ((i * point[0]) + (j * point[1]));

                    if(!((tpoint[0] < 0) || (tpoint[0] >= m_image->get_width()) || 
                        (tpoint[1] < 0) || (tpoint[1] >= m_image->get_height())))
                    {
                        sampler.SampleImage(tmp_img, tpoint[0], tpoint[1], bg, col);

                        p[0]+=col[0];
                        p[1]+=col[1];
                        p[2]+=col[2];
                        (*wp)+=1.0;
                    }

                    p+=3;
                    wp++;
                }
            }

            delete tmp_img;
        }
    }
    return added;
}

void 
s3d_merge::AverageAlign::PostProcess(void)
{
    size_t size = m_image->get_width() * m_image->get_height();
    float *p = m_image->get_buffer();
    float *wp = m_weights->get_buffer();

    for(size_t i=0;i<size;i++)
    {
        (*p)/=(*wp);
        p++;
        (*p)/=(*wp);
        p++;
        (*p)/=(*wp);
        p++;
        *wp = 1.0;
        wp++;
    }
}

s3d_merge::FloatImage *
s3d_merge::AverageAlign::MakeFloatImage(const Image16 *img)
{
    FloatImage *result(NULL);

    if(img)
    {
        result = new FloatImage(img->get_width(), img->get_height(), 3);

        if((m_dark) &&
           (m_dark->get_width() == result->get_width()) &&
           (m_dark->get_height() == result->get_height()))
        {
            CRUNCHD(result, img, m_dark, px_set_value);
        }
        else
        {
            CRUNCH(result, img, px_set_value);
        }
    }
    return result;
}
